"""Heater-Shaker Module sub-state."""
from dataclasses import dataclass
from typing import NewType, Optional, List


AbsorbanceReaderId = NewType("AbsorbanceReaderId", str)


@dataclass(frozen=True)
class AbsorbanceReaderSubState:
    """Absorbance-Plate-Reader-specific state."""

    module_id: AbsorbanceReaderId
    configured: bool
    measured: bool
    data: Optional[List[float]]
    configured_wavelength: Optional[int]



"""Common configuration command base models."""

from dataclasses import dataclass
from opentrons.hardware_control.nozzle_manager import (
    NozzleMap,
)
from ..resources import pipette_data_provider


@dataclass
class PipetteConfigUpdateResultMixin:
    """A mixin-suitable model for adding pipette config to private results."""

    pipette_id: str
    serial_number: str
    config: pipette_data_provider.LoadedStaticPipetteData


@dataclass
class PipetteNozzleLayoutResultMixin:
    """A nozzle layout result for updating the pipette state."""

    pipette_id: str

    nozzle_map: NozzleMap
    """A dataclass object holding information about the current nozzle configuration."""



"""Pick up tip command request, result, and implementation models."""
from __future__ import annotations
from dataclasses import dataclass
from opentrons_shared_data.errors import ErrorCodes
from pydantic import Field
from typing import TYPE_CHECKING, Optional, Type, Union
from typing_extensions import Literal

from opentrons.protocol_engine.errors.exceptions import TipNotAttachedError

from ..errors import ErrorOccurrence
from ..resources import ModelUtils
from ..types import DeckPoint
from .pipetting_common import (
    PipetteIdMixin,
    WellLocationMixin,
    DestinationPositionResult,
)
from .command import (
    AbstractCommandImpl,
    BaseCommand,
    BaseCommandCreate,
    DefinedErrorData,
    SuccessData,
)

if TYPE_CHECKING:
    from ..state import StateView
    from ..execution import MovementHandler, TipHandler


PickUpTipCommandType = Literal["pickUpTip"]


class PickUpTipParams(PipetteIdMixin, WellLocationMixin):
    """Payload needed to move a pipette to a specific well."""

    pass


class PickUpTipResult(DestinationPositionResult):
    """Result data from the execution of a PickUpTip."""

    # Tip volume has a default ONLY for parsing data from earlier versions, which did not include this in the result
    tipVolume: float = Field(
        0,
        description="Maximum volume of liquid that the picked up tip can hold, in µL.",
        ge=0,
    )

    tipLength: float = Field(
        0,
        description="The length of the tip in mm.",
        ge=0,
    )

    tipDiameter: float = Field(
        0,
        description="The diameter of the tip in mm.",
        ge=0,
    )


class TipPhysicallyMissingError(ErrorOccurrence):
    """Returned when sensors determine that no tip was physically picked up.

    That space in the tip rack is marked internally as not having any tip,
    as if the tip were consumed by a pickup.

    The pipette will act as if no tip was picked up. So, you won't be able to aspirate
    anything, and movement commands will assume there is no tip hanging off the bottom
    of the pipette.
    """

    isDefined: bool = True
    errorType: Literal["tipPhysicallyMissing"] = "tipPhysicallyMissing"
    errorCode: str = ErrorCodes.TIP_PICKUP_FAILED.value.code
    detail: str = "No tip detected."


@dataclass(frozen=True)
class TipPhysicallyMissingErrorInternalData:
    """Internal-to-ProtocolEngine data about a TipPhysicallyMissingError."""

    pipette_id: str
    labware_id: str
    well_name: str


_ExecuteReturn = Union[
    SuccessData[PickUpTipResult, None],
    DefinedErrorData[TipPhysicallyMissingError, TipPhysicallyMissingErrorInternalData],
]


class PickUpTipImplementation(AbstractCommandImpl[PickUpTipParams, _ExecuteReturn]):
    """Pick up tip command implementation."""

    def __init__(
        self,
        state_view: StateView,
        tip_handler: TipHandler,
        model_utils: ModelUtils,
        movement: MovementHandler,
        **kwargs: object,
    ) -> None:
        self._state_view = state_view
        self._tip_handler = tip_handler
        self._model_utils = model_utils
        self._movement = movement

    async def execute(
        self, params: PickUpTipParams
    ) -> Union[SuccessData[PickUpTipResult, None], _ExecuteReturn]:
        """Move to and pick up a tip using the requested pipette."""
        pipette_id = params.pipetteId
        labware_id = params.labwareId
        well_name = params.wellName
        well_location = params.wellLocation

        position = await self._movement.move_to_well(
            pipette_id=pipette_id,
            labware_id=labware_id,
            well_name=well_name,
            well_location=well_location,
        )

        try:
            tip_geometry = await self._tip_handler.pick_up_tip(
                pipette_id=pipette_id,
                labware_id=labware_id,
                well_name=well_name,
            )
        except TipNotAttachedError as e:
            return DefinedErrorData(
                public=TipPhysicallyMissingError(
                    id=self._model_utils.generate_id(),
                    createdAt=self._model_utils.get_timestamp(),
                    wrappedErrors=[
                        ErrorOccurrence.from_failed(
                            id=self._model_utils.generate_id(),
                            createdAt=self._model_utils.get_timestamp(),
                            error=e,
                        )
                    ],
                ),
                private=TipPhysicallyMissingErrorInternalData(
                    pipette_id=pipette_id,
                    labware_id=labware_id,
                    well_name=well_name,
                ),
            )
        else:
            return SuccessData(
                public=PickUpTipResult(
                    tipVolume=tip_geometry.volume,
                    tipLength=tip_geometry.length,
                    tipDiameter=tip_geometry.diameter,
                    position=DeckPoint(x=position.x, y=position.y, z=position.z),
                ),
                private=None,
            )


class PickUpTip(
    BaseCommand[PickUpTipParams, PickUpTipResult, TipPhysicallyMissingError]
):
    """Pick up tip command model."""

    commandType: PickUpTipCommandType = "pickUpTip"
    params: PickUpTipParams
    result: Optional[PickUpTipResult]

    _ImplementationCls: Type[PickUpTipImplementation] = PickUpTipImplementation


class PickUpTipCreate(BaseCommandCreate[PickUpTipParams]):
    """Pick up tip command creation request model."""

    commandType: PickUpTipCommandType = "pickUpTip"
    params: PickUpTipParams

    _CommandCls: Type[PickUpTip] = PickUpTip



"""Command models to measure absorbance."""
from __future__ import annotations
from typing import List, Optional, TYPE_CHECKING
from typing_extensions import Literal, Type

from pydantic import BaseModel, Field

from ..command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from ...errors.error_occurrence import ErrorOccurrence

if TYPE_CHECKING:
    from opentrons.protocol_engine.state import StateView
    from opentrons.protocol_engine.execution import EquipmentHandler


MeasureAbsorbanceCommandType = Literal["absorbanceReader/measure"]


class MeasureAbsorbanceParams(BaseModel):
    """Input parameters for a single absorbance reading."""

    moduleId: str = Field(..., description="Unique ID of the Absorbance Reader.")
    sampleWavelength: int = Field(..., description="Sample wavelength in nm.")


class MeasureAbsorbanceResult(BaseModel):
    """Result data from running an aborbance reading."""

    # TODO: Transform this into a more complex model, such as a map with well names.
    data: Optional[List[float]] = Field(
        ..., min_items=96, max_items=96, description="Absorbance data points."
    )


class MeasureAbsorbanceImpl(
    AbstractCommandImpl[
        MeasureAbsorbanceParams, SuccessData[MeasureAbsorbanceResult, None]
    ]
):
    """Execution implementation of an Absorbance Reader measurement."""

    def __init__(
        self,
        state_view: StateView,
        equipment: EquipmentHandler,
        **unused_dependencies: object,
    ) -> None:
        self._state_view = state_view
        self._equipment = equipment

    async def execute(
        self, params: MeasureAbsorbanceParams
    ) -> SuccessData[MeasureAbsorbanceResult, None]:
        """Initiate a single absorbance measurement."""
        abs_reader_substate = self._state_view.modules.get_absorbance_reader_substate(
            module_id=params.moduleId
        )
        # Allow propagation of ModuleNotAttachedError.
        abs_reader = self._equipment.get_module_hardware_api(
            abs_reader_substate.module_id
        )

        if abs_reader is not None:
            result = await abs_reader.start_measure(wavelength=params.sampleWavelength)
            return SuccessData(
                public=MeasureAbsorbanceResult(data=result),
                private=None,
            )

        return SuccessData(
            public=MeasureAbsorbanceResult(data=None),
            private=None,
        )


class MeasureAbsorbance(
    BaseCommand[MeasureAbsorbanceParams, MeasureAbsorbanceResult, ErrorOccurrence]
):
    """A command to execute an Absorbance Reader measurement."""

    commandType: MeasureAbsorbanceCommandType = "absorbanceReader/measure"
    params: MeasureAbsorbanceParams
    result: Optional[MeasureAbsorbanceResult]

    _ImplementationCls: Type[MeasureAbsorbanceImpl] = MeasureAbsorbanceImpl


class MeasureAbsorbanceCreate(BaseCommandCreate[MeasureAbsorbanceParams]):
    """A request to execute an Absorbance Reader measurement."""

    commandType: MeasureAbsorbanceCommandType = "absorbanceReader/measure"
    params: MeasureAbsorbanceParams

    _CommandCls: Type[MeasureAbsorbance] = MeasureAbsorbance



"""Command models for Absorbance Reader commands."""

from .initialize import (
    InitializeCommandType,
    InitializeParams,
    InitializeResult,
    Initialize,
    InitializeCreate,
)

from .measure import (
    MeasureAbsorbanceCommandType,
    MeasureAbsorbanceParams,
    MeasureAbsorbanceResult,
    MeasureAbsorbance,
    MeasureAbsorbanceCreate,
)

__all__ = [
    # absorbanace_reader/initialize
    "InitializeCommandType",
    "InitializeParams",
    "InitializeResult",
    "Initialize",
    "InitializeCreate",
    # absorbanace_reader/measure
    "MeasureAbsorbanceCommandType",
    "MeasureAbsorbanceParams",
    "MeasureAbsorbanceResult",
    "MeasureAbsorbance",
    "MeasureAbsorbanceCreate",
]



from abc import ABC, abstractmethod
from typing import Dict, List
from opentrons.drivers.types import (
    AbsorbanceReaderLidStatus,
    AbsorbanceReaderDeviceState,
)


class AbstractAbsorbanceReaderDriver(ABC):
    @abstractmethod
    async def connect(self) -> None:
        """Connect to absorbance reader"""
        ...

    @abstractmethod
    async def disconnect(self) -> None:
        """Disconnect from absorbance reader"""
        ...

    @abstractmethod
    async def is_connected(self) -> bool:
        """Check connection to absorbance reader"""
        ...

    @abstractmethod
    async def get_lid_status(self) -> AbsorbanceReaderLidStatus:
        ...

    @abstractmethod
    async def get_available_wavelengths(self) -> List[int]:
        ...

    @abstractmethod
    async def get_single_measurement(self, wavelength: int) -> List[float]:
        ...

    @abstractmethod
    async def initialize_measurement(self, wavelength: int) -> None:
        ...

    @abstractmethod
    async def get_status(self) -> AbsorbanceReaderDeviceState:
        ...

    @abstractmethod
    async def get_device_info(self) -> Dict[str, str]:
        """Get device info"""
        ...
